#!/usr/bin/env python3
from PIL import Image
import sys, math, re
ncol=0
dither=0
args=sys.argv[1:]
fname=None
csrc=False
outname="img.seq"
rlecompress = False
outimg=None
bgalgo=0
def helpme(s):
    if s:
        print(s)
    print("""
Usage: %s [options] filename
Options are:
-d        - dither (Floyd-Steinberg)
-n <nr>   - limit colors (2..16)
-p <cols> - use specific palette (hexadecimal digits)
-g        - same as -p 01bcf (grayscale palette)
-b <0..2> - background algorithm (default 0)
-o <name> - output file name (default img.seq)
-r        - compress data
-c        - create colors.asm and bitmap.asm
-i <name> - output image file name (default none)
-h        - this help
Options -n, -g and -p are mutually exclusive""" % sys.argv[0])
    exit(0)

def mkpal(a):
    if not re.match(r'[0-9a-fA-F]{2,}$',a):
        helpme('Bad palette')
    b=sorted(set(int(x,16) for x in a))
    if len(b) != len(a):
        helpme('Bad palette')
    return b

palette = None

while len(args) > 0:
    arg=args.pop(0)
    if arg == '-d':
        dither=1
        continue
    if arg == "-h":
        helpme(None)
    if arg == "-b":
        if len(arg) == 0:
            helpme("Missing argument")
        try:
            n=int(args.pop(0))
        except:
            helpme("Bad background algorithm")
        if n < 0 or n > 2:
            helpme("Bad background algorithm")
        bgalgo=n
        continue
    if arg == '-p':
        if palette is not None or ncol != 0:
            helpme('Illegal argument combination')
        if len(arg) == 0:
            helpme("Missing argument")
        palette = mkpal(args.pop(0))
        continue
    if arg == '-g':
        if palette is not None or ncol != 0:
            helpme('Illegal argument combination')
        palette=[0,1,11,12,15]
        continue
    if arg == "-o":
        if palette is not None or ncol != 0:
            helpme('Illegal argument combination')
        if len(arg) == 0:
            helpme("Missing argument")
        outname=args.pop(0)
        continue
    if arg == '-n':        
        if len(args) == 0:
            helpme("Missing argument")
        try:
            ncol=int(args.pop(0))
        except:
            helpme("Illegal argument")
        if ncol < 2 or ncol > 16:
            helpme("Use 2 to 16 colors")
        continue
    if arg == "-c":
        csrc=True
        continue
    if arg == "-r":
        rlecompress=True
        continue
    if arg == '-i':
        if len(args) == 0:
            helpme("Missing argument")
        outimg=args.pop(0)
        continue
    fname=arg
    break
if fname is None:
    helpme("Missing filename")
if args:
    helpme("Too many arguments")
if ncol == 0:
    ncol = 16

cpal=[(0,0,0),(255,255,255),(136,0,0),
    (170,255,238),(204,68,204),(0,204,85),
    (0,0,170),(238,238,119),(221,136,85),(102,68,0),(255,119,119),
    (51,51,51),(119,119,119),(170,255,102),(0,136,255),(187,187,187)]

dpal=[]
if palette:
    for n in palette:
        dpal.extend(cpal[n])
else:
    if ncol == 0:
        ncol = 16
    for n in range(ncol):
        dpal.extend(cpal[n])

image=Image.open(fname)
image=image.resize((160,200))
pal=Image.new('P',(16,16))
pal.putpalette((dpal*160)[:3*256])
im2=image.quantize(palette=pal, dither=dither)

cdif={}

def difcol(c1,c2):
    global cdif, cpal
    def sq(a):
        return a*a
    d=cdif.get((c1,c2),None)
    if d is not None:
        return d
    d=math.sqrt(
        sq(cpal[c1][0]-cpal[c2][0]) +
        sq (cpal[c1][1]-cpal[c2][1]) +
        sq(cpal[c1][2]-cpal[c2][2]))
    cdif[(c1,c2)]=d
    return d

def bests(clrs,p):
    mx=99999
    cl=0
    for n in clrs:
        d=difcol(n,p)
        if d < mx:
            mx=d
            cl=n
    return cl
    
def blockcolors(x0,y0):
    global im2
    rcl={}
    for x in range(4):
        for y in range(8):
            c=im2.getpixel((x+x0,y+y0))
            rcl[c] = rcl.get(c,0)+1
    rcl=list(rcl.items())
    rcl.sort(key=lambda x:-x[1])
    return rcl
    
def bestbg():
    global bgalgo
    cidx={}
    def oldbg():
        global im2
        for x in range(160):
            for y in range(200):
                p=im2.getpixel((x,y)) % ncol
                cidx[p]=cidx.get(p,0)+1
    if bgalgo == 0:
        oldbg()
    else:
        for x in range(0,160,4):
            for y in range(0,200,8):
                cls=blockcolors(x,y)
                if len(cls) <= 3:
                    continue
                for r in cls:
                    ad = 1 if bgalgo == 2 else r[1]
                    cidx[r[0]] = cidx.get(r[0],0)+r[1]
        if len(cidx) == 0:
            oldbg
    cidx=list(cidx.items())
    cidx.sort(key=lambda x:-x[1])
    return cidx[0][0]


bgcolor=bestbg()

def bestcol(x0,y0):
    global im2,bgcolor,ncol
    inx={}
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p == bgcolor:
                continue
            inx[p]=inx.get(p,0)+1
    inx=list(inx.items())
    inx.sort(key=lambda x:-x[1])
    colors=[bgcolor]
    while(len(colors) < 4 and len(inx) > 0):
        colors.append(inx.pop(0)[0])
    while(len(colors) < 4) :
        colors.append(bgcolor)
    if len(inx) == 0:
        return colors
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p in colors:
                continue
            p=bests(colors,p)
            im2.putpixel((x0+x,y0+y),p)
    return colors
    
colormap=[]
            
for y in range(0,200,8):
    for x in range(0,160,4):        
        color=bestcol(x,y)
        colormap.append(color)

def pixid(x,y):
    global im2,colormap,ncol
    idx = x // 4
    idy = y // 8
    cmap=colormap[idy*40+idx]
    p=im2.getpixel((x,y)) % ncol
    n=cmap.index(p)
    return n

def pixchar(x0,y0):
    px=[]
    for y in range(8):
        p=0
        for x in range(4):
            p = (p << 2) | pixid(x+x0, y+y0)
        px.append(p)
    return px


def rle(b):
    b=list(b)
    n=0
    lnx=0
    outarr=[]
    while len(b) > 0:
        c=b.pop(0)
        if len(b) == 0:
            lnx += 2
            break
        if b[0] == c:
            b.pop(0)
            t=2
            while len(b) > 0 and b[0] == c and t < 127:
                t += 1
                b.pop(0)
            outarr.append(t)
            outarr.append(c)
            #print(t)
            lnx += 2
            continue
        oar=[c]
        t=1
        while len(b) > 0 and t <127:
            if len(b) >=3 and b[0] ==b[1] and b[0] == b[2]:
                break
            oar.append(b.pop(0))
            t += 1
        outarr.append(t | 0x80)
        outarr.extend(oar)
        lnx += t
    outarr.append(0)
    return outarr


def genasm():
    global colormap, bgcolor, csrc,outname, rlecompress
    if rlecompress:
        seqf=[bgcolor | 0x80]
    else:
        seqf=[bgcolor]
    cmap=list(x[3] for x in colormap)
    if rlecompress:
        cmbp=rle(cmap)
        seqf.extend(cmbp)
    else:
        seqf.extend(cmap)
    lines=['colormap:']
    while(len(cmap) > 0):
        cms=cmap[:8]
        cmap=cmap[8:]
        byts=', '.join(list(str(x) for x in cms))
        lines.append('.byte '+byts)
    lines.append('screenmap:')
    cmap=list((x[1] << 4) | x[2] for x in colormap)
    if rlecompress:
        cmbp=rle(cmap)
        seqf.extend(cmbp)
    else:
        seqf.extend(cmap)
    while(len(cmap) > 0):
        cms=cmap[:8]
        cmap=cmap[8:]
        byts=', '.join(list(str(x) for x in cms))
        lines.append('.byte '+byts)
    lines.append('bgcolor .byte %d' % bgcolor)
    if csrc:
        f=open('colors.asm','w')
        w='\n'.join(lines)
        f.write(w+'\n')
        f.close()

    lines=[]
    pax=[]
    for y in range(0,200,8):
        for x in range(0,160,4):
            px=pixchar(x,y)
            pax.extend(px)
            line='.byte '+ (', '.join(str(n) for n in px))
            lines.append(line)
    if csrc:
        f=open('bitmap.asm','w')
        lines='\n'.join(lines)
        f.write(lines+'\n')
        f.close()
    if rlecompress:
        pax=rle(pax)
    
    seqf.extend(pax)
    bts=bytes(seqf)
    f=open(outname,'wb')
    f.write(bts)
    f.close()
    return

genasm()
im2=im2.resize((320,200))
if outimg:
    try:
        im2.save(outimg)
    except:
        im2.convert('RGB').save(outimg)
im2.show('')
